#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <c10/cuda/CUDAGuard.h>

// Each warp computes one output element using warp-level reduction
// This design avoids atomic operations since each warp exclusively works on one output element.
__global__ void linear_relu_unrolled_warp_kernel(const float* x, const float* weight, const float* bias, float* out,
                                                 int batch_size, int in_features, int out_features) {
  // Compute a global warp id from the thread index
  int warp_id = (blockIdx.x * blockDim.x + threadIdx.x) / 32;
  int lane = threadIdx.x % 32; // Lane index within the warp
  
  int total_outputs = batch_size * out_features;
  if (warp_id >= total_outputs) return;

  // Map warp id to output matrix coordinates (i, j)
  int i = warp_id / out_features;
  int j = warp_id % out_features;

  float sum = 0.0f;

  // Cache the base indices to avoid repeated calculations
  const int x_base = i * in_features;
  const int w_base = j * in_features;
  
  // Each thread in the warp processes a strided portion of the in_features dimension
  // Unroll by 4 to reduce loop overhead and enable better instruction-level parallelism
  #pragma unroll 4
  for (int k = lane; k < in_features; k += 32) {
    float x_val = x[x_base + k];
    float w_val = weight[w_base + k];
    sum += x_val * w_val;
  }

  // Perform warp-level reduction using shuffle operations
  for (int offset = 16; offset > 0; offset /= 2) {
    sum += __shfl_down_sync(0xffffffff, sum, offset);
  }

  // The first lane of each warp writes the result
  if (lane == 0) {
    sum += bias[j];
    // Apply ReLU activation
    out[i * out_features + j] = sum > 0.0f ? sum : 0.0f;
  }
}


torch::Tensor linear_relu_forward(torch::Tensor x, torch::Tensor weight, torch::Tensor bias) {
  TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
  TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
  TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

  const int batch_size = x.size(0);
  const int in_features = x.size(1);
  const int out_features = weight.size(0);

  // Allocate output tensor
  auto out = torch::empty({batch_size, out_features}, x.options());

  // Each warp computes one output element. Total number of warps required is batch_size * out_features.
  int total_warps = batch_size * out_features;

  // Each warp consists of 32 threads. Determine total threads required.
  int total_threads = total_warps * 32;

  // Choose block size as a multiple of 32, e.g., 256 threads per block
  int threads_per_block = 256;
  int blocks = (total_threads + threads_per_block - 1) / threads_per_block;

  cudaStream_t stream = c10::cuda::getCurrentCUDAStream();
  linear_relu_unrolled_warp_kernel<<<blocks, threads_per_block, 0, stream>>>(
      x.data_ptr<float>(),
      weight.data_ptr<float>(),
      bias.data_ptr<float>(),
      out.data_ptr<float>(),
      batch_size,
      in_features,
      out_features
  );
  
  return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
  m.def("forward", &linear_relu_forward, "Unrolled warp-level reduction GEMM+Bias+ReLU (CUDA)");
}
